@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2002-2016, University of Amsterdam
+    Copyright (c)  2002-2020, University of Amsterdam
                               VU University Amsterdam
+                              CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -285,7 +286,7 @@ user_agent('SWI-Prolog').
 %     * status_code(-Code)
 %     If this option is  present  and   Code  unifies  with the HTTP
 %     status code, do *not* translate errors (4xx, 5xx) into an
-%     exception. Instead, http_open/3 behaves as if 200 (success) is
+%     exception. Instead, http_open/3 behaves as if 2xx (success) is
 %     returned, providing the application to read the error document
 %     from the returned stream.
 %
@@ -361,24 +362,29 @@ user_agent('SWI-Prolog').
 %                         ])
 %               ==
 %
-%       @error existence_error(url, Id)
-%       @see ssl_context/3 for SSL related options if
-%       library(http/http_ssl_plugin) is loaded.
+%   @throws error(existence_error(url, Id),Context) is raised if the
+%   HTTP result code is not in the range 200..299. Context has the
+%   shape context(Message, status(Code, TextCode)), where `Code` is the
+%   numeric HTTP code and `TextCode` is the textual description thereof
+%   provided by the server. `Message` may provide additional details or
+%   may be unbound.
+%
+%   @see ssl_context/3 for SSL related options if
+%   library(http/http_ssl_plugin) is loaded.
 
 :- multifile
     socket:proxy_for_url/3.           % +URL, +Host, -ProxyList
 
 http_open(URL, Stream, QOptions) :-
-    meta_options(is_meta, QOptions, Options),
+    meta_options(is_meta, QOptions, Options0),
     (   atomic(URL)
     ->  parse_url_ex(URL, Parts)
     ;   Parts = URL
     ),
     autoload_https(Parts),
+    upgrade_ssl_options(Parts, Options0, Options),
     add_authorization(Parts, Options, Options1),
-    findall(HostOptions,
-            http:open_options(Parts, HostOptions),
-            AllHostOptions),
+    findall(HostOptions, hooked_options(Parts, HostOptions), AllHostOptions),
     foldl(merge_options_rev, AllHostOptions, Options1, Options2),
     (   option(bypass_proxy(true), Options)
     ->  try_http_proxy(direct, Parts, Stream, Options2)
@@ -476,6 +482,21 @@ http:http_connection_over_proxy(socks(SocksHost, SocksPort), _Parts, Host:Port,
             throw(Error)
           )).
 
+%!  hooked_options(+Parts, -Options) is nondet.
+%
+%   Calls  http:open_options/2  and  if  necessary    upgrades  old  SSL
+%   cacerts_file(File) option to a cacerts(List) option to ensure proper
+%   merging of options.
+
+hooked_options(Parts, Options) :-
+    http:open_options(Parts, Options0),
+    upgrade_ssl_options(Parts, Options0, Options).
+
+upgrade_ssl_options(Parts, Options0, Options) :-
+    (   requires_ssl(Parts)
+    ->  ssl:upgrade_legacy_options(Options0, Options)
+    ;   Options = Options0
+    ).
 
 merge_options_rev(Old, New, Merged) :-
     merge_options(New, Old, Merged).
@@ -498,14 +519,18 @@ host_and_port(Host, _,       Port,    Host:Port).
 %   If the requested scheme is https or wss, load the HTTPS plugin.
 
 autoload_https(Parts) :-
+    requires_ssl(Parts),
     memberchk(scheme(S), Parts),
-    secure_scheme(S),
     \+ clause(http:http_protocol_hook(S, _, StreamPair, StreamPair, _),_),
     exists_source(library(http/http_ssl_plugin)),
     !,
     use_module(library(http/http_ssl_plugin)).
 autoload_https(_).
 
+requires_ssl(Parts) :-
+    memberchk(scheme(S), Parts),
+    secure_scheme(S).
+
 secure_scheme(https).
 secure_scheme(wss).
 
@@ -669,12 +694,12 @@ add_method(Options0, Options) :-
     Options = [method(post)|Options0].
 add_method(Options0, [method(get)|Options0]).
 
-
 %!  do_open(+HTTPVersion, +HTTPStatusCode, +HTTPStatusComment, +Header,
 %!          +Options, +Parts, +Host, +In, -FinalIn) is det.
 %
-%   Handle the HTTP status. If 200, we   are ok. If a redirect, redo
-%   the open, returning a new stream. Else issue an error.
+%   Handle the HTTP status once available. If   200-299, we are ok. If a
+%   redirect, redo the open,  returning  a   new  stream.  Else issue an
+%   error.
 %
 %   @error  existence_error(url, URL)
 
@@ -717,7 +742,7 @@ do_open(Version, Code, _, Lines, Options, Parts, Host, In0, In) :-
     (   option(status_code(Code), Options),
         Lines \== []
     ->  true
-    ;   Code == 200
+    ;   successful_code(Code)
     ),
     !,
     parts_uri(Parts, URI),
@@ -746,6 +771,9 @@ do_open(_Version, Code, Comment, _,  _, Parts, _, _, _) :-
     throw(error(Formal, context(_, status(Code, Comment)))).
 
 
+successful_code(Code) :-
+    between(200, 299, Code).
+
 %!  redirect_limit_exceeded(+Options:list(compound), -Max:nonneg) is semidet.
 %
 %   True if we have exceeded the maximum redirection length (default 10).
@@ -1184,6 +1212,7 @@ add_authorization(_, Options, Options) :-
 add_authorization(Parts, Options0, Options) :-
     url_part(user(User), Parts),
     url_part(password(Passwd), Parts),
+    !,
     Options = [authorization(basic(User,Passwd))|Options0].
 add_authorization(Parts, Options0, Options) :-
     stored_authorization(_, _) ->   % quick test to avoid work
@@ -1419,22 +1448,31 @@ keep_alive(Connection) :-
 
 :- public keep_alive/4.
 
+keep_alive(StreamPair, Host, _In, 0) :-
+    !,
+    debug(http(connection), 'Adding connection to ~p to pool', [Host]),
+    add_to_pool(Host, StreamPair).
 keep_alive(StreamPair, Host, In, Left) :-
+    Left < 100,
+    debug(http(connection), 'Reading ~D left bytes', [Left]),
     read_incomplete(In, Left),
     add_to_pool(Host, StreamPair),
     !.
 keep_alive(StreamPair, _, _, _) :-
-    close(StreamPair, [force(true)]).
+    debug(http(connection),
+          'Closing connection due to excessive unprocessed input', []),
+    (   debugging(http(connection))
+    ->  catch(close(StreamPair), E,
+              print_message(warning, E))
+    ;   close(StreamPair, [force(true)])
+    ).
 
 %!  read_incomplete(+In, +Left) is semidet.
 %
 %   If we have not all input from  a Keep-alive connection, read the
 %   remainder if it is short. Else, we fail and close the stream.
 
-read_incomplete(_, 0) :- !.
 read_incomplete(In, Left) :-
-    Left < 100,
-    !,
     catch(setup_call_cleanup(
               open_null_stream(Null),
               copy_stream_data(In, Null, Left),
