@@ -829,14 +829,14 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
 #endif
       if ( k == ATOM_table_space )
       { if ( !LD->tabling.node_pool )
-	  LD->tabling.node_pool = new_alloc_pool("table_space", i);
+	  LD->tabling.node_pool = new_alloc_pool("private_table_space", i);
 	else
 	  LD->tabling.node_pool->limit = (size_t)i;
       }
 #ifdef O_PLMT
       else if ( k == ATOM_shared_table_space )
       { if ( !GD->tabling.node_pool )
-	{ alloc_pool *pool = new_alloc_pool("table_space", i);
+	{ alloc_pool *pool = new_alloc_pool("shared_table_space", i);
 	  if ( pool && !COMPARE_AND_SWAP(&GD->tabling.node_pool, NULL, pool) )
 	    free_alloc_pool(pool);
 	} else
@@ -1359,6 +1359,9 @@ initPrologFlags(void)
   setPrologFlag("agc_margin",FT_INTEGER,	       GD->atoms.margin);
 #endif
   setPrologFlag("table_space", FT_INTEGER, GD->options.tableSpace);
+#ifdef O_PLMT
+  setPrologFlag("shared_table_space", FT_INTEGER, GD->options.sharedTableSpace);
+#endif
   setPrologFlag("stack_limit", FT_INTEGER, LD->stacks.limit);
 #if defined(HAVE_DLOPEN) || defined(HAVE_SHL_LOAD) || defined(EMULATE_DLOPEN)
   setPrologFlag("open_shared_object",	  FT_BOOL|FF_READONLY, TRUE, 0);
