@@ -93,9 +93,12 @@ zown(zipper *z)
 
 static void
 zdisown(zipper *z)
-{ int tid = PL_thread_self();
-
+{
+#ifndef NDEBUG
+  int tid = PL_thread_self();
   assert(z->owner == tid);
+#endif
+
   if ( z->lock_count == 0 )
   { z->owner = 0;
     simpleMutexUnlock(&z->lock);
@@ -204,15 +207,28 @@ zrelease(zipper *z)
 		 *  ACCESS ARCHIVES AS STREAMS  *
 		 *******************************/
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+(*) As we map the resource file, we should  not write it in place. If we
+do so anyway, a Prolog process running on   this  state will crash if it
+tries to access its resources.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 static voidpf
 zopen64_file(voidpf opaque, const void* filename, int mode)
-{ char modes[4];
+{ char modes[8];
   char *m = modes;
 
   if ( (mode&ZLIB_FILEFUNC_MODE_CREATE) )
-    *m++ = 'w';
-  else
-    *m++ = 'r';
+  { *m++ = 'w';
+    strcpy(m, "m444");
+    m += strlen(m);
+    if ( ExistsFile(filename) )
+    { if ( !RemoveFile(filename) )
+	return NULL;			/* see (*) */
+    }
+  } else
+  { *m++ = 'r';
+  }
   *m++ = 'b';
   *m = EOS;
 
@@ -431,6 +447,7 @@ static int
 close_zipper(zipper *z)
 { zipFile zf;
   unzFile uzf;
+  char *path;
   int rc = 0;
 
   if ( (zf=z->writer) )
@@ -440,15 +457,19 @@ close_zipper(zipper *z)
   { z->reader = NULL;
     rc = unzClose(uzf);
   }
-  if ( z->path )
-  { free((char*)z->path);
-    z->path = NULL;
+  if ( (path=(char*)z->path) )
+  { z->path = NULL;
+    free(path);
   }
   if ( z->input.any )
   { switch(z->input_type)
     { case ZIP_STREAM:
 	if ( true(z, ZIP_CLOSE_STREAM_ON_CLOSE) )
-	  Sclose(z->input.stream);
+	{ IOSTREAM *in = z->input.stream;
+
+	  z->input.stream = NULL;
+	  Sclose(in);
+	}
         break;
       default:
 	break;
@@ -851,8 +872,9 @@ PRED_IMPL("zipper_open_new_file_in_zip", 4, zipper_open_new_file_in_zip, 0)
        PL_get_chars(A2, &fname, flags) &&
        zacquire(z, ZIP_WRITE_ENTRY, NULL, "new_file") )
   { int rc;
-    zip_fileinfo zipfi = {0};
+    zip_fileinfo zipfi;
 
+    memset(&zipfi, 0, sizeof(zipfi));
     zset_time(&zipfi, ftime);
     rc = zipOpenNewFileInZip4_64(z->writer, fname,
 				 &zipfi,
@@ -919,21 +941,25 @@ PRED_IMPL("zipper_goto", 2, zipper_goto, 0)
       { zrelease(z);
 	return FALSE;
       }
+      Sdprintf("zipper_goto/2: rc=%d (%s)\n", rc,
+	       a == ATOM_first ? "first" : "next");
       assert(0);
     } else if ( PL_is_functor(A2, FUNCTOR_file1) )
     { term_t arg = PL_new_term_ref();
       char *fname;
       int flags = (CVT_ATOM|CVT_STRING|CVT_EXCEPTION|REP_UTF8);
+      int rc;
 
       if ( PL_get_arg(1, A2, arg) &&
 	   PL_get_chars(arg, &fname, flags) )
-      { switch(unzLocateFile(z->reader, fname, TRUE))
+      { switch((rc=unzLocateFile(z->reader, fname, TRUE)))
 	{ case UNZ_OK:
 	    return TRUE;
 	  case UNZ_END_OF_LIST_OF_FILE:
 	    zrelease(z);
 	    return FALSE;
 	}
+	Sdprintf("zipper_goto/2: rc=%d file(%s)\n", rc, fname);
 	assert(0);
       }
     } else if ( PL_is_functor(A2, FUNCTOR_offset1) )
@@ -1319,8 +1345,9 @@ SopenZIP(zipper *z, const char *name, int flags)
     }
   } else
   { int rc;
-    zip_fileinfo zipfi = {0};
+    zip_fileinfo zipfi;
 
+    memset(&zipfi, 0, sizeof(zipfi));
     if ( !zacquire(z, ZIP_WRITE_ENTRY, NULL, "new_file") )
       return NULL;
 
